from datetime import datetime, timedelta
from typing import Any, Callable, Dict, List

import typer  # type: ignore
from google.auth.exceptions import RefreshError  # type: ignore

from .. import auth
from .date_utils import create_date_range


def create_body_list(
    body: Dict[Any, Any], new_body: List[Dict[Any, Any]] = [], granularity: str = None,
) -> List[Dict[Any, Any]]:
    """
    Gets a body, and creates a new body from that.
    """
    dates = create_date_range(
        start=body.get("startDate"), end=body.get("endDate"), granularity=granularity,
    )
    if granularity is not None:

        if granularity in [
            "monday",
            "tuesday",
            "wednesday",
            "thursday",
            "friday",
            "saturday",
            "sunday",
        ]:
            for idx in range(len(dates)):
                try:
                    """
                    If the detail level is low, medium or high
                    reduce end day by one to prevent duplications.
                    """
                    body.update({"startDate": dates[idx], "endDate": dates[idx]})
                    new_body.append(body.copy())
                except IndexError:
                    pass

        else:
            for idx in range(len(dates)):
                try:
                    """
                    If the detail level is low, medium or high
                    reduce end day by one to prevent duplications.
                    """
                    body.update(
                        {
                            "startDate": dates[idx],
                            "endDate": (
                                datetime.strptime(dates[idx + 1], "%Y-%m-%d")
                                - timedelta(days=1)
                            ).strftime("%Y-%m-%d"),
                        }
                    )
                    new_body.append(body.copy())
                except IndexError:
                    pass
    else:
        for date in dates:
            body.update({"startDate": date, "endDate": date})
            new_body.append(body.copy())

    return new_body


def path_exists(filename: str) -> bool:
    """
    Checks for the given file path exists.
    """
    from pathlib import Path

    return Path(filename).exists()


def regenerate_credentials(method: Callable) -> Callable:
    """
    If query raises RefreshError(Expired token causes this.)
    It automatically regenerate credentials
    and runs the method again.
    """

    def run_query(*args, **kw):
        try:
            result = method(*args, **kw)
        except RefreshError:
            regen = typer.confirm(
                "Your credentials has expired, do you want to regenerate them?",
                default=True,
            )

            if regen is True:
                auth.get_authenticated()
                typer.secho(
                    "Authenticated successfully!",
                    fg=typer.colors.BRIGHT_GREEN,
                    bold=True,
                ),
                result = method(*args, **kw)
            else:
                msg = typer.secho("Aborting...", bold=True, fg=typer.colors.BRIGHT_RED,)
                exit()

        return result

    return run_query



from typing import List

import typer  # type: ignore

from . import auth
from .utils.completion_utils import dimensions, export_type, month_complete, searchtype
from .utils.date_utils import (
    create_date,
    days_last_util,
    get_day_granularity,
    get_today,
    process_date,
)
from .utils.path_utils import create_toml_list
from .utils.query_utils import query_builder, query_deleter, query_lister
from .utils.selector_utils import create_granularity_selector, create_selector

app = typer.Typer(add_completion=False, name="Seoman")
query_app = typer.Typer(
    add_completion=False, name="Seoman", short_help="Create and run your queries."
)
app.add_typer(query_app, name="query")


@app.command("auth")
def get_auth():
    """
    Get authenticated.
    """

    auth.get_authenticated()


@app.command("manual")
def manual(
    url: str = typer.Option(..., help="That's how world wide web works."),
    start_date: str = typer.Option(
        None, help="Give a date to start [Example: 2020-03, 2 months ago, 4 ay önce]"
    ),
    end_date: str = typer.Option(
        None,
        help="Give a date to end [Example: 2020-08-11, 23 hours ago, yaklaşık 23 saat önce]",
    ),
    search_type: str = typer.Option(
        None, help="Search type ", autocompletion=searchtype
    ),
    dimensions: List[str] = typer.Option(
        None, help="Add dimensions ", autocompletion=dimensions
    ),
    row_limit: int = typer.Option(
        None, help="Maximum number of rows to return [Default is 1,000] "
    ),
    start_row: int = typer.Option(None, help="Set starting index [Default is 0]"),
    export: str = typer.Option(
        None, help="Specify export type ", autocompletion=export_type,
    ),
    granularity: str = typer.Option(
        None,
        help="Set a frequency/granularity or group your queries [Example: daily, twodaily, monday, tuesday, weekdays, weekends]",
    ),
):

    """
    Top pages in the site
    """
    service = auth.load_service()

    if start_date is not None:
        start = process_date(dt=start_date, which_date="start")
    if end_date is not None:
        end = process_date(dt=end_date, which_date="end")

    service.update_body({"startDate": start, "endDate": end or get_today()})

    if len(dimensions) >= 1:
        if "all" in dimensions:
            service.update_body(
                {"dimensions": ["date", "page", "query", "country", "device"]}
            )
        else:
            service.update_body({"dimensions": [dimension for dimension in dimensions]})

    if search_type is not None:
        service.update_body({"searchType": search_type})

    if row_limit is not None:
        service.update_body({"rowLimit": row_limit})

    if start_row is not None:
        service.update_body({"startRow": start_row})

    service.concurrent_query_asyncio(url=url, granularity=granularity or "daily")
    service.export(export_type=export, url=url, command="manual")


@app.command("sites")
def show_sites(
    url: str = typer.Option(None, help="Retrieve information about specific site."),
    export: str = typer.Option(
        None, help="Specify export type ", autocompletion=export_type,
    ),
):
    """
    List all the web sites or the permission level for the specific site that associated with the account.
    """

    service = auth.load_service()

    service.sites(url=url or None)

    service.export(export_type=export, command="sites", url=url or "seoman")


@app.command("sitemaps")
def show_sitemap(
    url: str = typer.Option(
        ..., help="The site's URL [Example: http://www.example.com/]"
    ),
    feedpath: str = typer.Option(
        None,
        help="The URL of the actual sitemap [Example: http://www.example.com/sitemap.xml]",
    ),
    export: str = typer.Option(
        None, help="Specify export type ", autocompletion=export_type,
    ),
):
    """
    List sitemaps-entries or get specific the sitemaps.
    """

    service = auth.load_service()

    service.sitemaps(url=url or None, feedpath=feedpath or None)

    service.export(export_type=export, command="sitemaps", url=url or "seoman")


@app.command("traffic")
def show_traffic(
    days: int = typer.Argument(None, help="Get total traffic from last given days."),
    site: str = typer.Option(
        None, help="Get traffic from specific website instead of all."
    ),
    export: str = typer.Option(
        None, help="Specify export type.", autocompletion=export_type,
    ),
):
    """
    Get total traffic from your sites.
    """

    service = auth.load_service()

    days = days or 30
    start, end = (
        days_last_util(days).get("startDate"),
        days_last_util(days).get("endDate"),
    )
    service.get_traffic(site=site or None, days=days)

    service.export(
        export_type=export or "table", command="traffic", url=f"{start}-{end}"
    )


@app.command("feedback")
def give_feedback():
    """
    Give us a feedback about your experience.
    """

    import webbrowser

    link = "https://github.com/zeoagency/seoman/issues/new"
    try:
        webbrowser.open(url=link)
    except webbrowser.Error:
        typer.secho(
            f"We could not find the browser, you can give us a feedback by visiting this link: {link}",
            bold=True,
            fg=typer.colors.RED,
        )


@app.command("version")
def show_version():
    """
    Show version and exit.
    """

    import pkg_resources

    typer.echo(pkg_resources.get_distribution("seoman").version)


@query_app.command("run")
def run_query(
    url: str = typer.Argument(
        None, help="Enter a url to override default one(If there is one.)"
    )
):
    """
    Select a query then run it.
    """
    name = create_selector(
        key="name", message="Select a query.", choices=create_toml_list()
    )
    service = auth.load_service()
    service.process_toml(filename=name)

    start_date, end_date = (
        service.__dict__["body"]["startDate"],
        service.__dict__["body"]["endDate"],
    )
    granularity = create_granularity_selector(start=start_date, end=end_date)

    try:
        toml_url, export_type = (
            service.__dict__["utils"]["url"],
            service.__dict__["utils"]["export-type"],
        )

    except KeyError:
        typer.secho(
            f"An error occured: Make sure you have export-type and url in your {name} file.",
            bold=True,
            fg=typer.colors.RED,
        )
        exit()

    service.concurrent_query_asyncio(
        url=url if url is not None else toml_url, granularity=granularity or "daily"
    )

    service.export(
        export_type=export_type, url=url if url is not None else toml_url, command=name
    )


@query_app.command("add")
def add_query():
    """
    Build a query interactively.
    """

    name = query_builder()

    _run = typer.confirm(f"{name} created successfully. Do you want to run it?")

    if _run:
        service = auth.load_service()
        service.process_toml(filename=name)
        start_date, end_date = (
            service.__dict__["body"]["startDate"],
            service.__dict__["body"]["endDate"],
        )
        granularity = create_granularity_selector(start=start_date, end=end_date)

        try:
            toml_url, export_type = (
                service.__dict__["utils"]["url"],
                service.__dict__["utils"]["export-type"],
            )

        except KeyError:
            typer.secho(
                f"An error occured: Make sure you have export-type and url in your {name} file.",
                bold=True,
                fg=typer.colors.RED,
            )
            exit()

        service.concurrent_query_asyncio(
            url=toml_url, granularity=granularity or "daily"
        )

        service.export(export_type=export_type, url=toml_url, command=name)


@query_app.command("delete")
def delete_query():
    """
    Delete a query.
    """

    selected = create_selector(
        key="toml_name", message="Select a query to delete.", choices=create_toml_list()
    )

    query_deleter(filename=selected)
    return


@query_app.command("show")
def show_query():
    """
    Show details from selected query.
    """

    selected = create_selector(
        key="toml_name",
        message="Select a query to list details.",
        choices=create_toml_list(),
    )
    query_lister(filename=selected)



from datetime import date, datetime
from time import strftime, strptime
from typing import Any, List, Union

import inquirer  # type: ignore
import typer  # type: ignore

from .config_utils import ALL_GRANULARITIES


def create_selector(key: str, message: str, choices: List) -> Any:
    """
    Generic function that creates a dropdown selector from any list.
    """

    questions = [inquirer.List(name=key, message=message, choices=choices)]
    answer = inquirer.prompt(questions)
    return answer[key]


def create_granularity_selector(start: Union[str, date, datetime], end: Union[str, date, datetime]) -> str:
    """Create a dropdown selector for granularity & frequency."""

    start, end = (
        strftime("%d %b %Y", strptime(start, "%Y-%m-%d")
                 if isinstance(start, str) else start.timetuple()),
        strftime("%d %b %Y", strptime(end, "%Y-%m-%d")
                 if isinstance(end, str) else end.timetuple()),
    )

    questions = [
        inquirer.List(
            name="granularity",
            message=f"How frequently you would like to get your data between {start} and {end}",
            choices=ALL_GRANULARITIES,
        )
    ]

    answer = inquirer.prompt(questions)
    return answer["granularity"]



import json
import sys
from datetime import date, datetime, timedelta
from typing import Dict, List, Optional, Union
import dateparser
import requests
import typer  # type: ignore


def get_start_date(days: int) -> str:
    """
    Create a start and end day. If days is 60 it will return timedelta(days=60) - datetime.today()
    """
    return str(datetime.today().date() - timedelta(days=days))


def get_today() -> str:
    """
    Get todays date.
    """
    return str(datetime.today().date())


def days_last_util(days: int) -> Dict[str, str]:
    """
    Create a start and end day.
    """
    startDate = get_start_date(days)
    endDate = get_today()
    return {"startDate": startDate, "endDate": endDate}


def create_date(year: int = None, month: int = None) -> str:
    """
    Create a datetime from given year and month if both params are None, use datetime.year.
    """
    if year and month:
        return f"{year}-{month}-01"
    if year:
        return f"{year}-01-01"
    if month:
        return f"{datetime.now().year}-{month}-01"
    else:
        return f"{datetime.now().year}-01-01"


def granularity_days(
    granularity: Union[str, None], start: Union[str, date], end: Union[str, date]
) -> List[str]:
    """
    Create weekdays or weekends, or specific days from given granularity.
    """
    if isinstance(start, str):
        start = datetime.strptime(start, "%Y-%m-%d").date()
    if isinstance(end, str):
        end = datetime.strptime(end, "%Y-%m-%d").date()

    # If granularity is
    if granularity in [1, 2, 3, 4, 5, 6, 7]:
        return [
            (start + timedelta(days=x)).strftime("%Y-%m-%d")
            for x in range((end - start).days + 1)
            if (start + timedelta(days=x)).isoweekday() == granularity
        ]

    elif granularity == 67:
        dates = [
            (start + timedelta(days=x)).strftime("%Y-%m-%d")
            for x in range((end - start).days + 1)
            if (start + timedelta(days=x)).isoweekday() == 6
            or (start + timedelta(days=x)).isoweekday() == 7
        ]
        if datetime.strptime(dates[0], "%Y-%m-%d").isoweekday() == 7:
            dates.insert(
                0,
                (
                    (
                        datetime.strptime(
                            dates[0], "%Y-%m-%d") - timedelta(days=1)
                    ).strftime("%Y-%m-%d")
                ),
            )
        if datetime.strptime(dates[-1], "%Y-%m-%d").isoweekday() == 6:
            dates.insert(
                0,
                (
                    (
                        datetime.strptime(
                            dates[0], "%Y-%m-%d") + timedelta(days=1)
                    ).strftime("%Y-%m-%d")
                ),
            )
        return dates

    elif granularity == 12345:
        dates = [
            (start + timedelta(days=x)).strftime("%Y-%m-%d")
            for x in range((end - start).days + 1)
            if (start + timedelta(days=x)).isoweekday() == 1
            or (start + timedelta(days=x)).isoweekday() == 5
        ]
        if datetime.strptime(dates[0], "%Y-%m-%d").isoweekday() == 5:
            dates.insert(
                0,
                (
                    (
                        datetime.strptime(
                            dates[0], "%Y-%m-%d") - timedelta(days=4)
                    ).strftime("%Y-%m-%d")
                ),
            )
        return dates

    elif granularity == 10:
        return [
            (start + timedelta(days=x)).strftime("%Y-%m-%d")
            for x in range((end - start).days + 1)
        ]


def create_date_range(
    days: Optional[int] = None,
    start: Optional[str] = None,
    end: Optional[str] = None,
    granularity: Optional[str] = None,
) -> List[str]:
    """
    Creates a list of dates from the given range and granularity
    Example: [2020-03-15, 2020-03-16,...]
    """

    if start is not None and end is not None:
        if granularity is not None:
            if get_weekday_by_name(granularity) != 10:
                granularity = get_weekday_by_name(granularity)  # type: ignore
                return granularity_days(granularity=granularity, start=start, end=end)
            else:
                day_interval = get_day_granularity(granularity)

        new_end, new_start = (
            datetime.strptime(
                end, "%Y-%m-%d").date() + timedelta(days=1),
            datetime.strptime(
                start, "%Y-%m-%d").date(),
        )

        day_diff = (new_end - new_start).days

        if day_diff < 0:
            typer.secho(
                f"There must be a problem with your start({new_start}) and end date({new_end}).",
                fg=typer.colors.GREEN,
                bold=True,
            )
            typer.secho(
                "Exiting..", fg=typer.colors.RED,
            )
            sys.exit()
        if day_interval > day_diff:
            typer.secho(
                f"Your date range is {day_diff}, it can not be smaller than your frequency which is {day_interval}.",
                fg=typer.colors.GREEN,
                bold=True,
            )
            typer.secho(
                "Exiting..", fg=typer.colors.RED,
            )
            sys.exit()
        dates = sorted(
            [
                (new_start + timedelta(days=x)).strftime("%Y-%m-%d")
                for x in range(0, abs(day_diff), day_interval)
            ]
        )
        if dates[0] != start:
            dates.insert(0, start)
        if dates[-1] != end:
            dates.append(end)
        return dates

    elif granularity is not None:
        start_date, end_date = create_date(year=2020), get_today()
        granularity = (
            get_weekday_by_name(granularity)  # type: ignore
            if get_weekday_by_name(granularity) != 10
            else get_day_granularity(granularity)
        )
        return granularity_days(granularity=granularity, start=start_date, end=end_date)

    elif days is not None:
        return sorted(
            [
                (datetime.today().date() - timedelta(days=x)).strftime("%Y-%m-%d")
                for x in range(days)
            ]
        )


def get_day_granularity(granularity: str) -> int:
    return {
        "daily": 1,
        "twodaily": 2,
        "threedaily": 3,
        "fourdaily": 4,
        "fivedaily": 5,
        "sixdaily": 6,
        "weekly": 7,
        "twoweekly": 14,
        "threeweekly": 21,
        "monthly": 30,
        "twomonthly": 60,
        "quarterly": 90,
        "yearly": 365,
    }.get(granularity, 1)


def get_weekday_by_name(day_name: str) -> int:
    return {
        "monday": 1,
        "tuesday": 2,
        "wednesday": 3,
        "thursday": 4,
        "friday": 5,
        "saturday": 6,
        "sunday": 7,
        "weekends": 67,
        "weekdays": 12345,
    }.get(day_name, 10)


def process_date(dt: str, which_date: str) -> Union[str, int]:
    """
    Process human readable datetime strings, to date objects then to str in %Y-%m-%d fmt.
    2 months ago -> datetime.date(2020, 7, 22) -> '2020-07-22'
    """
    retry = 0
    max_retry = 3
    parse = dt
    if dt:
        while retry < max_retry:
            try:
                new_dt = dateparser.parse(parse)
                if new_dt:
                    return new_dt.strftime("%Y-%m-%d")

            except Exception:
                typer.secho(
                    """There must be a problem with your date or the language is not supported yet.""",
                    bold=True,
                    fg=typer.colors.BRIGHT_RED,
                )
                typer.secho(
                    """Try entering in YYYY-MM-DD format.""",
                    bold=True,
                    fg=typer.colors.BRIGHT_RED,
                )
                enter_new = typer.confirm(
                    f"Do you want to enter a new {which_date} date?", default=True
                )
                if enter_new is False:
                    typer.secho(
                        "Aborting...", bold=True, fg=typer.colors.BRIGHT_RED,
                    )
                    sys.exit()
                else:
                    parse = typer.prompt("Enter here")

            finally:
                retry += 1

    typer.secho(
        """There must be a problem with your date or the language is not supported yet 😕,you might want to send a feedback with 'seoman feedback' then we can support your language.""",
        bold=True,
        fg=typer.colors.BRIGHT_RED,
    )
    sys.exit()



from typing import List


def searchtype() -> List[str]:
    return ["auto", "byPage", "byProperty"]


def dimensions() -> List[str]:
    return ["country", "query", "date", "device", "searchAppearance"]


def export_type() -> List[str]:
    return ["json", "csv", "table", "excel", "tsv"]


def month_complete() -> List[str]:
    return ["01", "02", "03", "04", "05", "06", "07", "08", "09", "10", "11", "12"]



from typing import List

ALL_GRANULARITIES: List[str] = [
    "daily",
    "twodaily",
    "threedaily",
    "fourdaily",
    "fivedaily",
    "sixdaily",
    "weekly",
    "twoweekly",
    "threeweekly",
    "monthly",
    "twomonthly",
    "quarterly",
    "yearly",
    "monday",
    "tuesday",
    "wednesday",
    "thursday",
    "friday",
    "saturday",
    "sunday",
    "weekends",
    "weekdays",
]



from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List

import inquirer  # type: ignore
import toml
import typer  # type: ignore

from seoman.utils.date_utils import process_date


def query_builder() -> str:
    typer.secho(
        """
    FORMATTING AND TIPS

    GENERAL FORMATTING
        [Tip] You are not supposed to answer questions if it is not [REQUIRED] 
        If you want to skip that question, just press space then enter.

    URL
        [Formatting: sc-domain:example.com or https://example.com]

    DATES
        [Formatting] Dates are in YYYY-MM-DD format.
        [Example] 23 march 2020 | 2020-03-10 | 2 weeks and 4 months ago 

    FILTERS
        [Formatting] If you want to add multiple filters split them by ',' 
        [Example] country equals FRA, device notContains tablet
        [Suggested Format] dimensions, operator, expression
    
    GRANULARITY
        Granularity specifies the frequency of the data, higher frequency means higher response time.
        [Examples] If you specify 'monday' seoman returns results only from mondays between start date and end date.
        [Examples] If you specify 'fivedaily' it splits your date range by 5 then runs unique queries.
        if your start date is 2020-03-10 and the end date is 2020-04-10 it first sends query for 03-10 to 03-15 then 03-15 to 03-20 then merges them all.   

    DIMENSIONS
        [Valid Parameters] page, query, date, device, country | for simplicity you can type 'all' to include all of them.
    
    EXPORT TYPE
        [Valid Parameters] excel, csv, json, tsv.

    ROW LIMIT
        [Valid Parameters] Must be a number from 1 to 25000.

    START ROW 
        [Valid Parameters] Must be a non-negative number.

    """,
        fg=typer.colors.BRIGHT_GREEN,
        bold=True,
    )

    questions = [
        inquirer.Text("url", message="[Required] The site's URL"),
        inquirer.Text(
            "start_date", message="[Required] Start date of the requested date range",
        ),
        inquirer.Text(
            "end_date", message="[Required] End date of the requested date range",
        ),
    ]

    answers = inquirer.prompt(questions)
    url = answers.get("url", "")
    start_date = answers.get("start_date", "")
    end_date = answers.get("end_date", "")

    questions = [
        inquirer.List(
            "dimensions",
            message="Which dimensions of Search Analytics you would like to group by?",
            choices=[
                "all [date, query, page, device, country]",
                "keywords & pages [date, query, page]",
                "by devices [date, device]",
                "by countries [date, countries]",
                "custom [Choose from: date - query - page - device - country]",
            ],
        ),
    ]

    answers = inquirer.prompt(questions)

    if (
        answers.get("dimensions")
        == "custom [Choose from: date - query - page - device - country]"
    ):
        questions = [
            inquirer.Checkbox(
                "dimensions",
                message="Which dimensions of Search Analytics you would like to group by?",
                choices=["date", "query", "page", "country", "device"],
            ),
        ]
        answers = inquirer.prompt(questions)
        dimensions = answers.get("dimensions", [])

    else:
        dimensions = answers.get("dimensions", "")

    questions = [
        inquirer.Text(
            "filters",
            message="Zero or more groups of filters to apply to the dimension grouping values",
        ),
        inquirer.Text(
            "start_row", message="First row of the response [Known as start-row]",
        ),
        inquirer.Text(
            "row_limit", message="The maximum number of rows to return [0-25000]",
        ),
        inquirer.List(
            "search_type",
            message="The search type to filter for",
            choices=["web", "image", "video"],
            default="web",
        ),
        inquirer.List(
            "export",
            message="The export type for the results",
            choices=["xlsx", "csv", "json", "tsv"],
        ),
    ]

    answers = inquirer.prompt(questions)
    filters = answers.get("filters", "")
    start_row = answers.get("start_row", "")
    row_limit = answers.get("row_limit", "")
    search_type = answers.get("search_type", "")
    export = answers.get("export", "")

    query: Dict[str, Dict[str, Any]] = {"query": {}}
    all_dimensions = ["page", "query", "date", "device", "country"]

    if len(url) > 5:
        query["query"].update({"url": url})

    if start_date.strip() != "":
        query["query"].update(
            {"start-date": process_date(dt=start_date, which_date="start")}
        )

    if end_date.strip() != "":
        query["query"].update({"end-date": process_date(dt=end_date, which_date="end")})

    if isinstance(dimensions, str):
        if dimensions == "all [date, query, page, device, country]":
            query["query"].update(
                {"dimensions": ["date", "query", "page", "device", "country"]}
            )

        elif dimensions == "keywords & pages [date, query, page]":
            query["query"].update({"dimensions": ["date", "query", "page"]},)

        elif dimensions == "by devices [date, device]":
            query["query"].update({"dimensions": ["date", "device"]},)

        elif dimensions == "by country [date, country]":
            query["query"].update({"dimensions": ["date", "country"]},)
    else:
        query["query"].update({"dimensions": [dim for dim in dimensions]})

    if filters.strip() != "":
        query["query"].update({"filters": [filt for filt in filters.split(",")]})

    if start_row.strip() != "" and start_row.isnumeric():
        query["query"].update({"start-row": start_row})

    if row_limit.strip() != "" and row_limit.isnumeric():
        if int(row_limit) >= 25000:
            row_limit = "25000"
        query["query"].update({"row-limit": row_limit.strip()})

    if search_type.strip() != "":
        query["query"].update({"search-type": search_type.strip().lower()})

    if export.strip() != "":
        query["query"].update({"export-type": export})

    typer.secho("\nYour query is ready\n", fg=typer.colors.BRIGHT_GREEN, bold=True)

    filename = typer.prompt("Give a name to your query") + ".toml"
    folder_path = Path.home() / ".queries"
    file_path = Path.home() / ".queries" / Path(filename)

    if not Path(folder_path).exists():
        Path(folder_path).mkdir(exist_ok=False)

    if not Path(file_path).exists():
        with open(file_path, "w") as file:
            toml.dump(query, file)

        return filename

    else:
        new_name = typer.prompt("File name already exists, enter a new name.") + ".toml"
        file_path = Path.home() / ".queries" / Path(new_name)

        if not Path(file_path).exists():
            with open(file_path, "w") as file:
                toml.dump(query, file)

    return new_name


def query_deleter(filename: str) -> None:
    """
    Delete a query from queries directory.
    """

    if not filename.endswith(".toml"):
        filename = filename + ".toml"

    p = Path.home() / ".queries" / filename

    p.unlink()


def query_lister(filename: str) -> None:
    """
    Show details of the selected query.
    """

    import toml
    from pytablewriter import UnicodeTableWriter  # type: ignore

    writer = UnicodeTableWriter()

    writer.table_name = filename

    if not filename.endswith(".toml"):
        filename = filename + ".toml"

    p = Path.home() / ".queries" / filename

    with open(str(p), "r") as file:
        query_file = toml.load(file)

    writer.headers = [k for k in query_file["query"]]
    writer.value_matrix = [
        [
            " ".join(v) if isinstance(v, list) else v
            for k, v in query_file["query"].items()
        ]
    ]

    writer.write_table()



import sys
from datetime import datetime
from typing import IO, Any, Dict, List, Optional, Tuple, Union

import typer  # type: ignore
from click import progressbar  # type: ignore

from .exceptions import FolderNotFoundError
from .utils.date_utils import create_date, days_last_util, get_today
from .utils.export_utils import Export
from .utils.service_utils import create_body_list, path_exists, regenerate_credentials


class SearchAnalytics:
    """
    Class that handles all the events in Seoman.
    """

    def __init__(self, service, credentials) -> None:
        self.service = service
        self.credentials = credentials
        self.data: Union[Dict[Any, Any], Any] = {}
        self.body: Dict[Any, Any] = {
            "startRow": 0,
            "rowLimit": 25000,
        }
        self.utils: Dict[str, str] = {}

    def update_body(self, body: Dict[Any, Any]) -> None:
        """
        Updates the body, that we are going to use in the query
        """

        self.body.update(**body)

    @regenerate_credentials
    def concurrent_query_asyncio(self, url: str, granularity: str = None) -> None:
        """
        Run queries concurrently.
        """

        import asyncio

        from googleapiclient.errors import HttpError  # type: ignore

        bodies = create_body_list(self.body, granularity=granularity)
        extra_bodies = []

        async def con_query(body, query_type: str) -> None:
            start_row = 25000
            try:
                data = (
                    self.service.searchanalytics()
                    .query(siteUrl=url, body=body)
                    .execute()
                )
            except HttpError:
                await asyncio.sleep(2)
                try:
                    data = (
                        self.service.searchanalytics()
                        .query(siteUrl=url, body=body)
                        .execute()
                    )
                except HttpError:
                    pass

            try:
                self.data.setdefault("rows", []).append(data["rows"])

                if len(data["rows"]) > 24999 and query_type == "first":
                    new_body = body.copy()
                    new_body.update({"startRow": 24999})
                    extra_bodies.append(new_body)

            except (KeyError, UnboundLocalError):
                pass

        async def main(
            body_list: List[Dict[Any, Any]], message: str, query_type: str
        ) -> None:
            with progressbar(
                body_list,
                label=message,
                length=len(body_list),
                fill_char="█",
                empty_char=" ",
            ) as bod:
                for body in bod:
                    await con_query(body, query_type=query_type)

        asyncio.run(
            main(body_list=bodies, message="Fetching data", query_type="first"))

        if len(extra_bodies) >= 1:
            confirm_rows = typer.confirm(
                f"More than 25.000 rows found for {len(extra_bodies)} query, do you want to include them too?"
            )
            if confirm_rows:
                asyncio.run(
                    main(
                        body_list=extra_bodies,
                        message="Fetching more data",
                        query_type="second",
                    )
                )

    @regenerate_credentials
    def sites(self, url: Union[None, str] = None) -> None:
        """
        List all the web sites associated with the account.

        Info: https://developers.google.com/resources/api-libraries/documentation/webmasters/v3/python/latest/webmasters_v3.sites.html
        """

        if url:
            self.data.update(self.service.sites().get(siteUrl=url).execute())

        else:
            self.data.update(
                {
                    key: [
                        v for v in value if v["permissionLevel"] != "siteUnverifiedUser"
                    ]
                    for key, value in self.service.sites().list().execute().items()
                }
            )

    @regenerate_credentials
    def get_traffic(self, site: str = None, days: int = 30) -> None:
        """
        Get your site's traffic results by given days. [Default: 30]
        """
        import asyncio

        if site is None:
            self.sites()

            async def con_query(url, idx) -> None:
                self.data["siteEntry"][idx].update(
                    {
                        "impression": self.service.searchanalytics()
                        .query(siteUrl=url, body=days_last_util(days=days))
                        .execute()["rows"][0]
                    }
                )

            async def main() -> None:
                with progressbar(
                    self.data["siteEntry"],
                    label="Fetching data",
                    length=len(self.data["siteEntry"]),
                    fill_char="█",
                    empty_char=" ",
                ) as dat:
                    for idx, data in enumerate(dat):
                        await con_query(data["siteUrl"], idx)

            asyncio.run(main())

            self.data = sorted(
                list(self.data.values())[0],
                key=lambda x: (
                    x["impression"]["clicks"],
                    x["impression"]["impressions"],
                ),
                reverse=True,  # Sort data by descending order
            )

        else:
            self.data.update(
                {
                    "site": site,
                    "impression": self.service.searchanalytics()
                    .query(siteUrl=site, body=days_last_util(days=days))
                    .execute()["rows"][0],
                }
            )

    @regenerate_credentials
    def sitemaps(self, url: str, feedpath: Union[None, str] = None) -> None:
        """
        Lists the sitemaps-entries submitted for the given site.

        Info: https://developers.google.com/resources/api-libraries/documentation/webmasters/v3/python/latest/webmasters_v3.sitemaps.html
        """

        if url and feedpath:
            self.data.update(
                self.service.sitemaps().get(siteUrl=url, feedpath=feedpath).execute()
            )

        elif url:
            self.data.update(
                self.service.sitemaps().list(siteUrl=url).execute())

    def export(
        self, command: str, export_type: Optional[str] = None, url: str = None,
    ) -> None:
        """
        Specify the export type.
        """

        export_types = ["csv", "json", "tsv", "table"]

        if self.data == {"rows": []}:
            typer.secho(
                "Results are empty. Make sure you have the entered url and you have rights to run it.",
                fg=typer.colors.RED,
                bold=True,
            )
            sys.exit()

        export_data = Export(self.data)

        if command == ("sites" or "sitemaps") and export_type is None:
            export_data.export_to_table()

        elif export_type in ["CSV", "JSON", "TSV", "TABLE"]:
            export_type = export_type.lower()

        elif export_type == "csv":
            export_data.export_to_csv(
                filename=self._create_filename(
                    url=url, command=command, filetype=export_type
                )
            )

        elif export_type == "json":
            export_data.export_to_json(
                filename=self._create_filename(
                    url=url, command=command, filetype=export_type
                )
            )

        elif export_type == "tsv":
            export_data.export_to_tsv(
                filename=self._create_filename(
                    url=url, command=command, filetype="tsv")
            )

        elif export_type == "table":
            export_data.export_to_table()

        elif (
            export_type == "excel"
            or export_type == "xlsx"
            or export_type not in export_types
            or export_type is None
        ):
            export_data.export_to_excel(
                filename=self._create_filename(
                    url=url, command=command, filetype="xlsx"
                )
            )

    def _create_filename(self, url: Optional[str], command: str, filetype: str) -> str:
        """
        Creates a file name from timestamp, url and command.
        """

        from datetime import datetime

        def __clean_url(url: Optional[str]) -> str:
            for t in (
                ("https", ""),
                ("http", ""),
                (":", ""),
                ("sc-domain", ""),
                ("//", ""),
                ("/", "-"),
                ("--", "-"),
                (".", "-"),
                (",", "-"),
            ):
                if url is not None:
                    url = url.lower().replace(*t)

            return url or "query"

        def __create_name(file_exists: bool = False) -> str:
            from random import randint

            if not file_exists:
                return "-".join(
                    [
                        __clean_url(url),
                        command,
                        datetime.now().strftime(
                            "%d-%B-%Y-%H-%M") + f".{filetype}",
                    ]
                )

            return "-".join(
                [
                    __clean_url(url) or "sites",
                    command,
                    f"report-{randint(1,10000)}",
                    datetime.now().strftime("%d-%B-%Y-%H-%M") + f".{filetype}",
                ]
            )

        return (
            __create_name()
            if not path_exists(__create_name())
            else __create_name(file_exists=True)
            if not path_exists(__create_name(file_exists=True))
            else __create_name(file_exists=True)
        )

    def process_toml(self, filename: str) -> None:
        """
        Open and prepare toml files for querying.
        """

        from pathlib import Path

        import toml

        queries_path = Path.home() / ".queries"

        if not Path(queries_path).exists():
            typer.secho(
                "We couldn't find queries folder, make sure you are in the same directory.",
                fg=typer.colors.RED,
                bold=True,
            )

        if not filename.endswith(".toml"):
            if "." not in filename:
                filename = filename + ".toml"

        file_path = queries_path / filename
        with file_path.open("r") as file:
            query_file = toml.load(file)

        for key, value in query_file["query"].items():

            if key == "dimensions":
                if "all" in value:
                    self.update_body(
                        {"dimensions": ["date", "page",
                                        "query", "country", "device"]}
                    )
                else:
                    self.body.update(
                        {"dimensions": [dimension for dimension in value]})

            if key == "filters":
                self.body.update(
                    {
                        "dimensionFilterGroups": [
                            {
                                "filters": [
                                    {
                                        # Country
                                        "dimension": filters.split()[0],
                                        # Equals
                                        "operator": filters.split()[1],
                                        "expression": " ".join(
                                            filters.split()[2:]
                                        ),  # FRA
                                    }
                                    for filters in value
                                ]
                            }
                        ]
                    }
                )

            if key == "start-date":
                self.body.update({"startDate": value})

            if key == "end-date":
                self.body.update({"endDate": value})

            if key == "start-date":
                self.body.update({"startDate": value})

            if key == "search-type":
                self.body.update({"searchType": value})

            if key == "row-limit":
                self.body.update({"rowLimit": value})

            if key == "start-row":
                self.body.update({"startRow": value})

            if key == "export-type":
                self.utils.update({"export-type": value})

            if key == "url":
                self.utils.update({"url": value})



class SeomanException(Exception):
    """
    Generic exception class for user-friendly errors for Seoman.
    """

    def __init__(self, message: str) -> None:
        self.message = message

    def __str__(self) -> str:
        class_name = self.__class__.__name__
        return f"{class_name}: {self.message}"


class BrokenFileError(SeomanException):
    def __init__(self, message: str) -> None:
        super().__init__(message)


class FolderNotFoundError(SeomanException):
    def __init__(self, message: str) -> None:
        super().__init__(message)


class MissingParameterError(SeomanException):
    def __init__(self, message: str) -> None:
        super().__init__(message)



import collections.abc
import json
import sys
from pathlib import Path
from typing import IO, Any, Dict, List, Optional, Union

import typer  # type: ignore
from apiclient import discovery  # type: ignore
from google.oauth2.credentials import Credentials  # type: ignore
from google_auth_oauthlib.flow import InstalledAppFlow  # type: ignore

from .exceptions import BrokenFileError
from .service import SearchAnalytics
from .utils import selector_utils


def authenticate(
    client_config: Union[str, Path] = None,
    credentials: Union[Any, Dict[str, Dict[str, str]]] = None,
    serialize: Union[str, Path] = None,
    flow: str = "web",
) -> SearchAnalytics:

    if not credentials:

        if isinstance(client_config, collections.abc.Mapping):

            auth_flow = InstalledAppFlow.from_client_config(
                client_config=client_config,
                scopes=["https://www.googleapis.com/auth/webmasters.readonly"],
            )

        elif isinstance(client_config, str):
            try:
                auth_flow = InstalledAppFlow.from_client_secrets_file(
                    client_secrets_file=client_config,
                    scopes=["https://www.googleapis.com/auth/webmasters.readonly"],
                )
            except FileNotFoundError:
                typer.secho(
                    "\nAuthentication failed ❌\nReason: client_secrets.json not found in your current directory.",
                    fg=typer.colors.RED,
                    bold=True,
                )
                sys.exit()

        else:
            raise BrokenFileError("Client secrets must be a mapping or path to file")

        if flow == "web":
            auth_flow.run_local_server()
        elif flow == "console":
            auth_flow.run_console()
        else:
            raise ValueError("Authentication flow '{}' not supported".format(flow))

        credentials = auth_flow.credentials

    else:

        if isinstance(credentials, str):
            try:
                with open(credentials, "r") as f:
                    credentials = json.load(f)
            except FileNotFoundError:
                typer.secho(
                    "\nAuthentication failed ❌\nReason: credentials.json not found in your current directory.",
                    fg=typer.colors.RED,
                    bold=True,
                )
                sys.exit()

        credentials = Credentials(
            token=credentials["token"],
            refresh_token=credentials["refresh_token"],
            id_token=credentials["id_token"],
            token_uri=credentials["token_uri"],
            client_id=credentials["client_id"],
            client_secret=credentials["client_secret"],
            scopes=credentials["scopes"],
        )

    service = discovery.build(
        serviceName="webmasters",
        version="v3",
        credentials=credentials,
        cache_discovery=False,
    )

    if serialize:

        if isinstance(serialize, str):

            serialized = {
                "token": credentials.token,
                "refresh_token": credentials.refresh_token,
                "id_token": credentials.id_token,
                "token_uri": credentials.token_uri,
                "client_id": credentials.client_id,
                "client_secret": credentials.client_secret,
                "scopes": credentials.scopes,
            }

            with open(serialize, "w") as f:
                json.dump(serialized, f, indent=4)

        else:
            raise TypeError("`serialize` must be a path.")

    return SearchAnalytics(service, credentials)


def load_service(
    credentials: Optional[str] = f"{Path.cwd()}/credentials.json",
) -> SearchAnalytics:

    service = authenticate(credentials=credentials)
    return service


def get_authenticated(
    client_config: Optional[str] = f"{Path.cwd()}/client_secrets.json",
    serialize: Optional[str] = f"{Path.cwd()}/credentials.json",
) -> SearchAnalytics:

    path = Path.cwd().glob("*.json")

    # Files comes like PosixPath("/some/path.toml") we are converting it to just filename without extension
    files: Union[List[Path], List[str]] = [
        str(file).split("/")[-1].replace(".json", "") for file in path if file.is_file()
    ]

    cred_files = [file for file in files if str(file).startswith("client_secret")]

    if len(cred_files) == 1:
        client_config = str(cred_files[0]) + ".json"
    elif len(cred_files) > 1:
        client_config = (
            selector_utils.create_selector(
                key="config",
                message="More than 1 file found that starts with client_secret, select one.",
                choices=cred_files,
            )
            + ".json"
        )
    else:
        typer.secho(
            "nAuthentication failed ❌\nReason: We could not find the client_secrets in your current directory"
        )

    service = authenticate(client_config=client_config, serialize=serialize)
    typer.secho("Authentication complete ✅\n", fg=typer.colors.GREEN, bold=True)
    typer.secho(
        "seoman is brought to you by https://seo.do - world's most powerful rank tracker tool made by technical SEO's for large companies and agencies.",
        fg=typer.colors.BRIGHT_BLUE,
        bold=True,
    )
    return service



import csv
import json
import sys
from collections import OrderedDict
from time import time
from typing import Any, Dict, List, Optional, Union

import typer  # type: ignore
from halo import Halo  # type: ignore
from pytablewriter import TsvTableWriter, UnicodeTableWriter  # type: ignore


class Export:
    def __init__(
        self, data: Dict[Any, Any] = {}, keys: List[Any] = [], values: List[Any] = []
    ) -> None:
        self.data = data
        self.keys = keys
        self.values = values

    def _flatten(self, data: Dict[Any, Any], sep="_") -> OrderedDict:

        obj = OrderedDict()

        def recurse(temp, parent_key=""):
            """
            Recursive iterator to reach everything inside data.
            """

            if isinstance(temp, list):
                for i in range(len(temp)):
                    recurse(
                        temp[i], parent_key + sep + str(i) if parent_key else str(i)
                    )
            elif isinstance(temp, dict):
                for key, value in temp.items():
                    recurse(value, parent_key + sep + key if parent_key else key)
            else:
                obj[parent_key] = temp

        recurse(data)
        return obj

    def _split_to_kv(self, data: Dict[Any, Any]) -> None:
        """
        Split data to key value pair.
        """
        if (
            data.keys().__contains__("path")
            or data.keys().__contains__("permissionLevel")
            or data.keys().__contains__("sitemap_0_path")
        ):
            """
            If data contains path, it must be a sitemap instance.
            If data contains permissionLevel, it must be sites instance.
            Else a SearchAnalytics report.
            """

            # TODO FIX SITEMAPS
            for key, value in data.items():

                if key.__contains__("_"):

                    key = key.split("_")

                    if key[0] == "sitemap":
                        if len(key) >= 4:
                            key = " ".join(
                                [key[2], key[4] if key[4] != "0" else "", key[3]]
                            )
                        else:
                            key = " ".join([key[2]])

                    else:
                        key = " ".join([key[0], key[2], key[1]])

                if key not in self.keys:
                    self.keys.append(key)

                self.values.append(value)

        else:

            for key, value in data.items():
                key = key.split("_")

                if key[0] == "responseAggregationType":
                    continue

                try:
                    if key[-2] == "keys":
                        # If there are multiple keys, add key number to keys eg. [keys1, keys2]
                        key = key[-2] + key[-1]
                except IndexError:
                    pass

                key = key[-1] if isinstance(key, list) else key

                if key not in self.keys:
                    self.keys.append(key)

                self.values.append(value)

    def __preprocess(self) -> None:
        """
        Preprocess the data.
        """

        self._split_to_kv(self._flatten(self.data))

    def export_to_table(self) -> None:
        """
        Export in Unicode Table format.
        """

        self.__preprocess()

        sub = len(self.keys)

        writer = UnicodeTableWriter()

        writer.table_name = "Analytics"
        writer.margin = 2

        writer.headers = self.keys

        if sub >= 1:
            writer.value_matrix = [
                self.values[ctr : ctr + sub] for ctr in range(0, len(self.values), sub)
            ]
        else:
            typer.secho(
                "An error occured please check your query.",
                fg=typer.colors.RED,
                bold=True,
            )
            sys.exit()

        writer.write_table()

    @Halo("Exporting to JSON", spinner="dots")
    def export_to_json(self, filename: str) -> None:
        """
        Export in JSON format.
        """

        with open(filename, "w") as file:
            json.dump(self.data, file, indent=4, ensure_ascii=False)

        print(f"Analytics successfully created in JSON format ✅")

    @Halo("Exporting to CSV", spinner="dots")
    def export_to_csv(self, filename: str) -> None:
        """
        Export in CSV format.
        """

        self.__preprocess()

        sub = len(self.keys)

        from csv import writer

        with open(filename, "w") as file:
            csv_writer = writer(file)
            csv_writer.writerow(self.keys)
            for ctr in range(0, len(self.values), sub):
                csv_writer.writerow(self.values[ctr : ctr + sub])

        typer.secho(
            "\nAnalytics successfully created in CSV format ✅", bold=True,
        )

    @Halo("Exporting to Excel", spinner="dots")
    def export_to_excel(self, filename: str) -> None:
        """
        Export in XLSX format.
        """
        from pyexcelerate import Workbook  # type: ignore

        self.__preprocess()

        sub = len(self.keys)

        if sub >= 1:
            data = [
                self.values[ctr : ctr + sub] for ctr in range(0, len(self.values), sub)
            ]
        else:
            typer.secho(
                "An error occured please check your query.",
                fg=typer.colors.RED,
                bold=True,
            )
            sys.exit()

        data.insert(0, self.keys)

        wb = Workbook()

        ws = wb.new_sheet("Analytics", data=data)

        wb.save(filename)

        typer.secho(
            "\nAnalytics successfully created in XLSX format ✅", bold=True,
        )

    @Halo("Exporting to TSV", spinner="dots")
    def export_to_tsv(self, filename: str) -> None:
        """
        Export in TSV format.
        """

        self.__preprocess()

        sub = len(self.keys)

        writer = TsvTableWriter()

        writer.headers = self.keys
        if sub >= 1:
            writer.value_matrix = [
                self.values[ctr : ctr + sub] for ctr in range(0, len(self.values), sub)
            ]
        else:
            typer.secho(
                "An error occured please check your query.",
                fg=typer.colors.RED,
                bold=True,
            )
            sys.exit()

        writer.dump(filename)
        typer.secho(
            "\nAnalytics successfully created in TSV format ✅", bold=True,
        )



from pathlib import Path
from typing import List, Union

import typer  # type: ignore


def create_toml_list() -> List[str]:
    """Create a list from all the files that ends with .toml"""
    path = Path(Path.home() / ".queries").glob("*.toml")
    files: Union[List[Path], List[str]] = [file for file in path if file.is_file()]

    # Files comes like PosixPath("/some/path.toml") we are converting it to just filename without extension
    files = [str(file).split("/")[-1].split(".")[0] for file in files]

    if len(files) >= 1:
        return files

    typer.secho("0 query found, add some.", bold=True, fg=typer.colors.BRIGHT_RED)
    exit()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['seoman', 'seoman.utils']

package_data = \
{'': ['*']}

install_requires = \
['dateparser>=1.1.0,<2.0.0',
 'google-api-python-client>=1.12.5,<2.0.0',
 'google-auth-oauthlib>=0.4.1,<0.5.0',
 'halo>=0.0.30,<0.0.31',
 'inquirer>=2.7.0,<3.0.0',
 'pyexcelerate>=0.9.0,<0.10.0',
 'pytablewriter>=0.58.0,<0.59.0',
 'requests>=2.24.0,<3.0.0',
 'toml>=0.10.1,<0.11.0',
 'typer>=0.3.2,<0.4.0']

entry_points = \
{'console_scripts': ['seoman = seoman.main:app']}

setup_kwargs = {
    'name': 'seoman',
    'version': '0.1.10',
    'description': 'A Command Line tool that is designed for technical SEOs',
    'long_description': None,
    'author': 'seo.do',
    'author_email': 'yagizcan@zeo.org',
    'maintainer': 'Yagiz Degirmenci',
    'maintainer_email': 'yagizcan@zeo.org',
    'url': 'https://seo.do/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
